#include "OptionAmerican.h"

//Empty allocator. Necessary to register it in the Factory
OptionPutAmericanSingle::OptionPutAmericanSingle()
{
  this->kind = "OptionPutAmericanSingle";
  this->steps = 1000;
  this->exercised = 0;
}

OptionPutAmericanSingle::OptionPutAmericanSingle(double S, double E,
                                                 double tau, double r,
                                                 double q, double sig)
{
  lattice lat = {0.0,0.0};

  this->kind = "OptionPutAmericanSingle";
  this->steps = 1000;
  this->exercised = 0;
  this->E = E;
  this->tau = tau;
  this->r = r;
  this->pushAsset(1,S,q,sig);
  //Memory allocation and initialization.
  this->nodes = new std::vector<lattice>(2*this->steps+1,lat);
  this->dt = this->tau/(double)this->steps;
}

OptionPutAmericanSingle::~OptionPutAmericanSingle(){}

void OptionPutAmericanSingle::set_props(int steps)
{
  lattice lat = {0.0,0.0};

  if(IS_ODD(steps)) this->steps = steps+1;
  if(IS_EVEN(steps)) this->steps = steps;
  if(steps < 2) std::cout << "nsteps must be even >2. Changed" <<
      std::endl;
  delete this->nodes;
  this->nodes = new std::vector<lattice>(2*this->steps+1,lat);
}

double OptionPutAmericanSingle::value()
{
  double S = this->assets->front().S;
  double E = this->E;
  double tau = this->tau;
  double r = this->r;
  double q = this->assets->front().q;
  double sig = this->assets->front().sig;

  double u = exp(sig*sqrt(this->dt));
  double d = exp(-sig*sqrt(this->dt));
  double a = exp((r-q)*this->dt);
  double p = (a-d)/(u-d);

  int n = this->steps;
  double payoff = 0.0;
  double european = 0.0;
  lattice last = {0.0,0.0};
  int j;

#pragma omp parallel for private(j,payoff)
  for(j = 0; j < 2*n+1; j +=2){
    payoff = S*exp((n-j)*sig*sqrt(this->dt));
    this->nodes->at(j).s = payoff;
    /*
      This computation just evaluates if it is more profitable
      to exercise the option or to hold the position.
      */
    this->nodes->at(j).v = MAX(E-payoff,0.0);
  }

  for(n = this->steps; n>1; n--){
#pragma omp parallel for private(j,payoff,european)
    for(j = 1; j < 2*n+1; j+=2){
      payoff = S*exp((n-j)*sig*sqrt(this->dt));
      this->nodes->at(j).s = payoff;
      european = exp(-r*this->dt)*(p*this->nodes->at(j-1).v+
                                  (1-p)*this->nodes->at(j+1).v);
      this->nodes->at(j).v = MAX(E-payoff,european);
    }
    for(j = 1; j < 2*n+1; j+=2){//Refresh nodes
      this->nodes->at(j-1).s = this->nodes->at(j).s;
      this->nodes->at(j-1).v = this->nodes->at(j).v;
    }
  }

  //The zero lattice is special
  last.s = S;
  european = exp(-r*this->dt)*(p*this->nodes->at(1).v+
                               (1-p)*this->nodes->at(3).v);
  last.v = MAX(E-last.s, european);

  //Important to compute the greeks
  this->f = last.v;
  this->exercised = 1;

  return last.v;
}


double OptionPutAmericanSingle::delta()
{
double spoil = this->value();
return (this->nodes->at(1).v-this->nodes->at(3).v)/
    (this->nodes->at(1).s-this->nodes->at(3).s);
}

double OptionPutAmericanSingle::gamma()
{
  double deltauv;
  double deltavw;
  double suw;

  double spoil = this->value();

  deltauv = (this->nodes->at(0).v-this->nodes->at(2).v)/
            (this->nodes->at(0).s-this->nodes->at(2).s);
  deltavw = (this->nodes->at(2).v-this->nodes->at(4).v)/
            (this->nodes->at(2).s-this->nodes->at(4).s);
  suw = 0.5*(this->nodes->at(0).s-this->nodes->at(4).s);
  return 2*(deltauv-deltavw)/suw;
}

/*********************************************
        OptionCallAmericanSingle
  ********************************************/

//Empty allocator. Necessary to register it in the Factory
OptionCallAmericanSingle::OptionCallAmericanSingle()
{
  this->kind = "OptionCallAmericanSingle";
  this->steps = 100;
  this->exercised = 0;
}

OptionCallAmericanSingle::OptionCallAmericanSingle(double S, double E,
                                                   double tau, double r,
                                                   double q, double sig)
{
  lattice lat = {0.0,0.0};

  this->kind = "OptionCallAmericanSingle";
  this->steps = 100;
  this->exercised = 0;
  this->E = E;
  this->tau = tau;
  this->r = r;
  this->pushAsset(1,S,q,sig);
  //Memory allocation and initialization.
  this->nodes = new std::vector<lattice>(2*this->steps+1,lat);
  this->dt = this->tau/(double)this->steps;
}

OptionCallAmericanSingle::~OptionCallAmericanSingle(){}

void OptionCallAmericanSingle::set_props(int steps)
{
  lattice lat = {0.0,0.0};

  if(IS_ODD(steps)) this->steps = steps+1;
  if(IS_EVEN(steps)) this->steps = steps;
  if(steps < 2) std::cout << "nsteps must be even >2. Changed" <<
      std::endl;
  delete this->nodes;
  this->nodes = new std::vector<lattice>(2*this->steps+1,lat);
}

double OptionCallAmericanSingle::value()
{
  double S = this->assets->front().S;
  double E = this->E;
  double tau = this->tau;
  double r = this->r;
  double q = this->assets->front().q;
  double sig = this->assets->front().sig;

  double u = exp(sig*sqrt(this->dt));
  double d = exp(-sig*sqrt(this->dt));
  double a = exp((r-q)*this->dt);
  double p = (a-d)/(u-d);

  int n = this->steps;
  double payoff = 0.0;
  double european = 0.0;
  lattice last = {0.0,0.0};
  int j;

#pragma omp parallel for private(j,payoff)
  for(j = 0; j < 2*n+1; j +=2){
    payoff = S*exp((n-j)*sig*sqrt(this->dt));
    this->nodes->at(j).s = payoff;
    /*
      This computation just evaluates if it is more profitable
      to exercise the option or to hold the position.
      */
    this->nodes->at(j).v = MAX(payoff-E,0.0);
  }

  for(n = this->steps; n>1; n--){
#pragma omp parallel for private(j,payoff,european)
    for(j = 1; j < 2*n+1; j+=2){
      payoff = S*exp((n-j)*sig*sqrt(this->dt));
      this->nodes->at(j).s = payoff;
      european = exp(-r*this->dt)*(p*this->nodes->at(j-1).v+
                                  (1-p)*this->nodes->at(j+1).v);
      this->nodes->at(j).v = MAX(payoff-E,european);
    }
    for(j = 1; j < 2*n+1; j+=2){//Refresh nodes
      this->nodes->at(j-1).s = this->nodes->at(j).s;
      this->nodes->at(j-1).v = this->nodes->at(j).v;
    }
  }

  //The zero lattice is special
  last.s = S;
  european = exp(-r*this->dt)*(p*this->nodes->at(1).v+
                               (1-p)*this->nodes->at(3).v);
  last.v = MAX(last.s-E, european);

  //Important to compute the greeks
  this->f = last.v;
  this->exercised = 1;

  return last.v;
}


double OptionCallAmericanSingle::delta()
{
  double spoil = this->value();
  return (this->nodes->at(1).v-this->nodes->at(3).v)/
      (this->nodes->at(1).s-this->nodes->at(3).s);

}

double OptionCallAmericanSingle::gamma()
{
  double deltauv;
  double deltavw;
  double suw;

  double spoil = this->value();

  deltauv = (this->nodes->at(0).v-this->nodes->at(2).v)/
            (this->nodes->at(0).s-this->nodes->at(2).s);
  deltavw = (this->nodes->at(2).v-this->nodes->at(4).v)/
            (this->nodes->at(2).s-this->nodes->at(4).s);
  suw = 0.5*(this->nodes->at(0).s-this->nodes->at(4).s);
  return 2*(deltauv-deltavw)/suw;

}
